@@ -88,10 +88,11 @@ def assert_flattened_data_as_expected(
     flattened_ds = TimeseriesFlattener(
         prediction_times_df=prediction_times_df,
         n_workers=4,
+        drop_pred_times_with_insufficient_look_distance=False,
     )
 
-    flattened_ds._add_temporal_col_to_flattened_dataset(  # pylint: disable=protected-access
-        output_spec=output_spec,
+    flattened_ds.add_spec(  # pylint: disable=protected-access
+        spec=output_spec,
     )
 
     if expected_df:
@@ -145,7 +146,13 @@ def synth_prediction_times():
     return load_synth_prediction_times()
 
 
+@pytest.fixture(scope="function")
+def synth_predictor():
+    """Load the synth outcome times."""
+    return load_synth_outcome(n_rows=1_000)
+
+
 @pytest.fixture(scope="function")
 def synth_outcome():
     """Load the synth outcome times."""
-    return load_synth_outcome()
+    return load_synth_outcome(n_rows=1_000)
